@@ -1,21 +1,18 @@
+import numpy as np
+import scipy.linalg
 import tensorly as tl
 import torch
 from tensorly.cp_tensor import cp_to_tensor
 from tensorly.decomposition import parafac
+from torch import diag
 
 
-def _minimal_regularisation(M, eps):
-    # calculate smallest ammount of regularisation that ensures smallest eigenvalue is eps
-    M_smallest_eig = torch.relu(-torch.min(torch.linalg.eigvalsh(M))) + eps
-    M = M + M_smallest_eig * torch.eye(M.shape[0], device=M.device)
-    return M
-
-
-def _compute_matrix_power(M, p):
-    # torch.linalg.eig can be unstable if eigenvalues are the same or are small https://pytorch.org/docs/stable/generated/torch.linalg.eig.html
-    U, V = torch.linalg.eigh(M)
-    M_p = V @ torch.diag(torch.pow(U, p)) @ V.T
-    return M_p
+def mat_pow(mat, pow_, epsilon):
+    # Computing matrix to the power of pow (pow can be negative as well)
+    [D, V] = torch.linalg.eigh(mat)
+    mat_pow = V @ diag((D + epsilon).pow(pow_)) @ V.T
+    mat_pow[mat_pow != mat_pow] = epsilon  # For stability
+    return mat_pow
 
 
 def _demean(*views):
@@ -53,13 +50,15 @@ class MCCA:
 
         # Get the block covariance matrix placing Xi^TX_i on the diagonal
         D = torch.block_diag(
-            *[(1 - self.r) * m.T @ m + self.r * torch.eye(m.shape[1], device=m.device) for i, m in enumerate(views)])
+            *[
+                (1 - self.r) * m.T @ m + self.r * torch.eye(m.shape[1], device=m.device)
+                for i, m in enumerate(views)
+            ]
+        )
 
         C = C - torch.block_diag(*[view.T @ view for view in views]) + D
 
-        D = _minimal_regularisation(D, self.eps)
-
-        R = torch.linalg.inv(torch.linalg.cholesky(D))
+        R = mat_pow(D, -0.5, self.eps)
 
         # In MCCA our eigenvalue problem Cv = lambda Dv
         C_whitened = R @ C @ R.T
@@ -69,7 +68,7 @@ class MCCA:
         # Sort eigenvalues so lviewest first
         idx = torch.argsort(eigvals, descending=True)
 
-        eigvals = eigvals[idx[:self.latent_dims]]
+        eigvals = eigvals[idx[: self.latent_dims]]
 
         # leaky relu encourages the gradient to be driven by positively correlated dimensions while also encouraging
         # dimensions associated with spurious negative correlations to become more positive
@@ -105,15 +104,19 @@ class GCCA:
         # H is n_views * n_samples * k
         views = _demean(*views)
 
-        eigen_views = [view @ torch.inverse(_minimal_regularisation(view.T @ view, self.eps)) @ view.T for view in
-                       views]
+        eigen_views = [
+            view
+            @ mat_pow(view.T @ view, -1, self.eps)
+            @ view.T
+            for view in views
+        ]
 
         Q = torch.stack(eigen_views, dim=0).sum(dim=0)
         eigvals = torch.linalg.eigvalsh(Q)
 
         idx = torch.argsort(eigvals, descending=True)
 
-        eigvals = eigvals[idx[:self.latent_dims]]
+        eigvals = eigvals[idx[: self.latent_dims]]
 
         # leaky relu encourages the gradient to be driven by positively correlated dimensions while also encouraging
         # dimensions associated with spurious negative correlations to become more positive
@@ -162,11 +165,15 @@ class CCA:
         H1bar, H2bar = _demean(H1, H2)
 
         SigmaHat12 = (1.0 / (n - 1)) * H1bar.T @ H2bar
-        SigmaHat11 = (1 - self.r) * (1.0 / (n - 1)) * H1bar.T @ H1bar + self.r * torch.eye(o1, device=H1.device)
-        SigmaHat22 = (1 - self.r) * (1.0 / (n - 1)) * H2bar.T @ H2bar + self.r * torch.eye(o2, device=H2.device)
+        SigmaHat11 = (1 - self.r) * (
+                1.0 / (n - 1)
+        ) * H1bar.T @ H1bar + self.r * torch.eye(o1, device=H1.device)
+        SigmaHat22 = (1 - self.r) * (
+                1.0 / (n - 1)
+        ) * H2bar.T @ H2bar + self.r * torch.eye(o2, device=H2.device)
 
-        SigmaHat11RootInv = _compute_matrix_power(_minimal_regularisation(SigmaHat11, self.eps), -0.5)
-        SigmaHat22RootInv = _compute_matrix_power(_minimal_regularisation(SigmaHat22, self.eps), -0.5)
+        SigmaHat11RootInv = mat_pow(SigmaHat11, -0.5, self.eps)
+        SigmaHat22RootInv = mat_pow(SigmaHat22, -0.5, self.eps)
 
         Tval = SigmaHat11RootInv @ SigmaHat12 @ SigmaHat22RootInv
         trace_TT = Tval.T @ Tval
@@ -202,20 +209,27 @@ class TCCA:
         m = z[0].size(0)
         z = [z_ - z_.mean(dim=0).unsqueeze(dim=0) for z_ in z]
         covs = [
-            (1 - self.r) * (1.0 / (m - 1)) * z_.T @ z_ + self.r * torch.eye(z_.size(1), device=z_.device)
-            for
-            z_ in z]
-        z = [z_ @ _compute_matrix_power(_minimal_regularisation(cov, self.eps), -0.5) for z_, cov in zip(z, covs)]
-
-        for i, el in enumerate(z):
+            (1 - self.r) * (1.0 / (m - 1)) * z_.T @ z_
+            + self.r * torch.eye(z_.size(1), device=z_.device)
+            for z_ in z
+        ]
+        whitened_z = [z_ @ mat_pow(cov, -0.5, self.eps) for z_, cov in zip(z, covs)]
+        # The idea here is to form a matrix with M dimensions one for each view where at index
+        # M[p_i,p_j,p_k...] we have the sum over n samples of the product of the pth feature of the
+        # ith, jth, kth view etc.
+        for i, el in enumerate(whitened_z):
+            # To achieve this we start with the first view so M is nxp.
             if i == 0:
-                curr = el
+                M = el
+            # For the remaining views we expand their dimensions to match M i.e. nx1x...x1xp
             else:
-                for _ in range(len(curr.size()) - 1):
+                for _ in range(len(M.size()) - 1):
                     el = torch.unsqueeze(el, 1)
-                curr = torch.unsqueeze(curr, -1) @ el
-        M = torch.mean(curr, 0)
-        tl.set_backend('pytorch')
-        M_parafac = parafac(M.detach(), self.latent_dims)
+                # Then we perform an outer product by expanding the dimensionality of M and
+                # outer product with the expanded el
+                M = torch.unsqueeze(M, -1) @ el
+        M = torch.mean(M, 0)
+        tl.set_backend("pytorch")
+        M_parafac = parafac(M.detach(), self.latent_dims, verbose=False)
         M_hat = cp_to_tensor(M_parafac)
         return torch.norm(M - M_hat)
